/*	
 *	frampky.c
 *
 *	P/ECE POKEY(CO12294) Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2004 Naoyuki Sawa
 *
 *	* Mon Jan 26 20:20:00 JST 2004 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *	֐
 ****************************************************************************/

int
pokey_process(POKEY* pokey, int cycle)
{
	/* LFSR */
	pokey_lfsr_process(pokey, cycle);

	/* Channel 1`4 */
	return pokey_channel_process(pokey, &pokey->channel[0], cycle)	/*  t5bit */
	     + pokey_channel_process(pokey, &pokey->channel[1], cycle)	/* +t5bit */
	     + pokey_channel_process(pokey, &pokey->channel[2], cycle)	/* +t5bit */
	     + pokey_channel_process(pokey, &pokey->channel[3], cycle);	/* +t5bit */
									/* =t7bit */
}

void
pokey_lfsr_process(POKEY* pokey, int cycle)
{
	POKEYLFSR* lfsr;
	const unsigned char* table;

	/* 4bit LFSR */
	lfsr = &pokey->lfsr4;
	table = pokey_lfsr_table4;
	lfsr->progress += cycle;
	lfsr->progress %= (1 << 4) - 1; /* (2^4)-1TCNŃ[v */
	lfsr->output = table[lfsr->progress / 8] >> (lfsr->progress & 7) & 1;

	/* 5bit LFSR */
	lfsr = &pokey->lfsr5;
	table = pokey_lfsr_table5;
	lfsr->progress += cycle;
	lfsr->progress %= (1 << 5) - 1; /* (2^5)-1TCNŃ[v */
	lfsr->output = table[lfsr->progress / 8] >> (lfsr->progress & 7) & 1;

	/* 17or9bit LFSR */
	lfsr = &pokey->lfsr17_9;
	if(pokey->reg[8] & 0x80) { /* 9bit */
		table = pokey_lfsr_table9;
		lfsr->progress += cycle;
		lfsr->progress %= (1 << 9) - 1; /* (2^9)-1TCNŃ[v */
		lfsr->output = table[lfsr->progress / 8] >> (lfsr->progress & 7) & 1;
	} else { /* 17bit */
		table = pokey_lfsr_table17;
		lfsr->progress += cycle;
		lfsr->progress %= (1 << 17) - 1; /* (2^17)-1TCNŃ[v */
		lfsr->output = table[lfsr->progress / 8] >> (lfsr->progress & 7) & 1;
	}
}

int
pokey_channel_process(POKEY* pokey, POKEYCHANNEL* channel, int cycle)
{
	if(channel->volume_only) return (channel->volume << 1) - 15; /* -15`15 */
	//̔菇ԈႦȂ悤!!
	//VolumeOnly[hłvolume=0́Ao0ł͂Ȃ-15ł!!
	if(!channel->volume) return 0;

	channel->progress += cycle;
	while(channel->progress >= 0) {
		channel->progress -= channel->period;
		switch(channel->distotion) {
		case 0: /* 5bit & 17bit */
			if(pokey->lfsr5.output && pokey->lfsr17_9.output) channel->output = !channel->output;
			break;
		case 1:
		case 3: /* 5bit */
			if(pokey->lfsr5.output                          ) channel->output = !channel->output;
			break;
		case 2: /* 5bit & 4bit */
			if(pokey->lfsr5.output && pokey->lfsr4.output   ) channel->output = !channel->output;
			break;
		case 4: /* 17bit */
			if(pokey->lfsr17_9.output                       ) channel->output = !channel->output;
			break;
		case 6: /* 4bit */
			if(pokey->lfsr4.output                          ) channel->output = !channel->output;
			break;
		default:/* case 5,7: Pure Tone */
			channel->output = !channel->output;
			break;
		}
	}

	return channel->output ?  channel->volume
			       : -channel->volume;
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

void
pokey_mix(POKEY* pokey, short wbuff[/*count*/], int count)
{
	DDA progress;
	int cycle;
	int sample;

	progress = pokey->progress; /* ̂߁Aꎞo */
	while(count > 0) {
		/* o߃TCN߂܂B */
		cycle = progress.u;
		progress.d += progress.n;
		if(progress.d >= 0) {
			progress.d -= SPEAKER_FREQUENCY;
			cycle++;
		}

		/* 1Tv~LVO܂B */
		sample = pokey_process(pokey, cycle); /* t7bit */
		sample <<= (16 - 7) + 1/**/;

		/* NbsOĊi[܂B */
		if(sample < -0x7fff) sample = -0x7fff;
		if(sample >  0x7fff) sample =  0x7fff;
		*wbuff++ = sample;

		/* cTv炵܂B */
		count--;
	}
	pokey->progress = progress; /* ߂BYꂸȂ!! */
}

